import os.path

import numpy as np
import xgboost as xgb


class XGBPredictor:
    def __init__(self, PATH, specify_anomaly=True):
        """
        Initiates the XGBoost model.
        :param PATH: path to provided model files
        :param binary: combine all anomaly labels into single ANOMALY
        """
        model_path = f'{PATH}/model.json'

        self.model = xgb.XGBClassifier()
        self.model.load_model(model_path)

        self.train_mean = np.load(f'{PATH}/train_mean.npy')
        self.train_std = np.load(f'{PATH}/train_std.npy')
        self.test_mean = np.load(f'{PATH}/test_mean.npy')
        self.test_std = np.load(f'{PATH}/test_std.npy')

        input_file = os.path.join(PATH, 'model_inputs.txt')
        if os.path.exists(input_file):
            with open(input_file, 'r') as f:
                input_values = f.readlines()
            ips = []
            for val in input_values:
                if 'log_delta_Turbidity' in val:
                    ips.append('logTurbidity')
                else:
                    ips.append(val.replace('\n', '').replace('delta_', ''))

            self.value_keys = ips
        else:
            self.value_keys = ['Temp', 'Turbidity', 'pH', 'ORP', 'Conductivity']

        self.standartize_idxs = {self.value_keys[i]: i for i in range(len(self.value_keys))}

        if specify_anomaly:
            all_labels = ['NORMAL', 'SURFACE_WATER', 'WASTEWATER', 'GROUNDWATER', 'DISCOLORATION', 'OTHER']
            self.label_to_idx = dict()
            for i in range(len(all_labels)):
                self.label_to_idx[all_labels[i]] = i
            self.idx_to_label = {self.label_to_idx[key]: key for key in self.label_to_idx.keys()}
        else:
            all_labels = ['NORMAL', 'ANOMALY', 'ANOMALY', 'ANOMALY', 'ANOMALY', 'ANOMALY']
            self.idx_to_label = {i: all_labels[i] for i in range(len(all_labels))}
            self.label_to_idx = {
                'NORMAL': 0,
                'ANOMALY': 1
            }

    def predict(self, t0: dict, t1: dict):
        """
        Predicts anomaly based from two datapoints with absolute values. Both have to be dictionaries of absolute values with ['Temp', 'Turbidity', 'pH', 'ORP' 'Conductivity'] keys present. Keys can contain lists of values, in which case output will be for each pair between t0 and t1.
        :param t0: datapoint in the past against which deltas will be calculated.
        :type t0: dict
        :param t1: current datapoint.
        :type t1: dict

        :returns results: list of predicted labels
        """
        # check if all necessary values present
        for key in self.value_keys:
            if not (key in t0.keys()):
                raise Exception(f'T0 does not contain absolute value with key {key}.')
            if not (key in t1.keys()):
                raise Exception(f'T0 does not contain absolute value with key {key}.')

        deltas = dict()
        for key in self.value_keys:
            delta = t1[key] - t0[key]
            deltas[key] = (delta - self.train_mean[self.standartize_idxs[key]]) / self.train_std[
                self.standartize_idxs[key]]

        vec = np.array([deltas[key] for key in self.value_keys]).T
        if len(vec.shape) == 1:
            vec = vec[np.newaxis, :]

        result = self.model.predict(vec)
        label_result = [self.idx_to_label[r] for r in result]
        return label_result

    def to_ints(self, rs):
        return [self.label_to_idx[r] for r in rs]
