from mysql.connector import Error
import pandas as pd
import numpy as np
import time
from datetime import datetime
from multiprocessing import Pool, cpu_count
import yaml
import sqlalchemy
from sqlalchemy import sql

from asya_modelis import XGBPredictor


def GetConnection(db_uri):
    # Change TAGs before using the script
    engine = sqlalchemy.create_engine(db_uri, poolclass=sqlalchemy.NullPool)
    return engine.connect()


def get_positives(row, hist_rec):
    res = np.argwhere(row.to_numpy() == 1).flatten().tolist()
    if len(res)>0:
        return ', '.join(str(num) for num in [hist_rec[i] for i in res])
    else:
        return ''


def make_decision(model_forecast: pd.DataFrame, hist_rec: list):
    ens_pos = model_forecast.apply(lambda row: get_positives(row, hist_rec), axis=1)
    model_forecast['prob'] = np.mean(model_forecast, axis=1)
    model_forecast['anom'] = 0
    model_forecast.loc[model_forecast['prob'] > 0.8, ['anom']] = 1
    model_forecast['ens_pos'] = ens_pos
    return model_forecast[['anom', 'prob', 'ens_pos']]
      
      
def getSQL(df, insert_sql):
    """
    Generates SQL script for inserting multiple rows
    Args:
        df (DataFrame): Dataframe with forecasted values.

    Returns:
        str: SQL query, that can be executed to insert forecasted values into database.
    """
    df.reset_index(inplace=True, drop=True)
    if len(df) > 0:
        sql = insert_sql
        template = " ('{}',{},'{}',{},{},'{}','{}'),"
        for i in range(len(df)):
            sql += template.format(df.loc[i, 'ID'], 
                                   df.loc[i, 'SensorSetID'], 
                                   df.loc[i, 'ModelID'], 
                                   df.loc[i, 'Anomaly'],
                                   df.loc[i, 'AnomalyProb'],
                                   df.loc[i, 'ModelParameters'],
                                   df.loc[i, 'Ensemble_Positives'])

        return sqlalchemy.text(sql[:-1])
    else:
        return "SELECT -1"
    

def get_forecast(sensorset=1):
    with open('config/config.yaml') as file:
        config= yaml.safe_load(file)
    
    # Column with reading date and time
    reading_datetime_col = config['parameters']['reading_datetime_col']
    
    # How many new records should be selected for one iteration
    max_new_records = config['parameters']['max_new_records']
    
    # How manu historical recirds should be taken. Depends on readings per second.
    # A list of 10 values is returned
    hist_records = config['parameters']['min_hist_records_ss1'] 
    
    min_hist_records = hist_records[0]
    min_hist_records_ens = np.max(hist_records)
    
    # Model configuration
    model_config = config['models']['ORP_TOC_WT4']
    modelid = model_config['modelid']
    modelid_ens = f'{modelid}_ens'
    
    # Query configuration
    query_config = config['queries']
    
    # DB connection string
    db_uri = config['parameters']['db_uri']

    # Query to get the historical data
    sql_get_historic_data = query_config['get_historical_data']
    
    # Query to get the new data
    sql_get_new_data = query_config['get_new_data']
    
    # Columns for making a forecast
    pred_cols = model_config['pred_cols'] 
    pred_cols_m = model_config['pred_cols']
    
    # Loading model
    model = XGBPredictor(PATH=model_config['path'], specify_anomaly=False)    
    # Stoppers
    stopper = 1
    stopper_ens = 0                                  
    while True:
        # Infinite loop
        try:
            if (stopper > 0) and (stopper_ens > 0):
                break
            
            # SINGLE MODEL PREDICTION
            # ******************************************************************************************************
            if stopper == 0:
                cn = GetConnection(db_uri)
                # Read new data (without the forecast)
                data = pd.read_sql(sql.text(sql_get_new_data.format(modelid,
                                                                    str(sensorset), 
                                                                    str(max_new_records))), cn)
                cn.close() 
                data.sort_values(by=[reading_datetime_col], ascending=False, inplace=True, ignore_index=True)
                l_new = len(data)
                if l_new > 0:
                    # New data is available
                    # Get the reading date and time of the first record
                    data[reading_datetime_col] = pd.to_datetime(data[reading_datetime_col])
                    dt = data[reading_datetime_col].min() 
                    # Get the minimal necessary amount of historical data to perform the forecast
                    cn = GetConnection(db_uri)
                    data_hist = pd.read_sql(sql.text(sql_get_historic_data.format(modelid,
                                                                                str(sensorset), 
                                                                                str(dt), 
                                                                                str(min_hist_records))), cn, )
                    cn.close() 
                    l_hist = len(data_hist) 
                    if l_hist >= min_hist_records:
                        # There is enough records to perform the forecast
                        # Concatenate historical and new data
                        data_full = pd.concat([data, data_hist], ignore_index=True)
                        # Sort in descending order by the reading date time so the last new reading would be the first one
                        data_full.sort_values(by=[reading_datetime_col], ascending=False, inplace=True, ignore_index=True) 
                        data['ModelID'] = modelid
                        data['ModelParameters'] = 'Single model at interval t-15 minutes' 
                        data['Ensemble_Positives'] = ''
                        
                        # Calculating the means
                        for c in pred_cols:
                            data_full[f'{c}_m'] = data_full[c].rolling(window=min_hist_records).mean()
                        
                        
                        # Read new records in the list
                        df_T1 = data_full.loc[0:l_new-1][pred_cols].copy()
                        T1 = df_T1.to_dict('list')
                        df_T0 = data_full.loc[min_hist_records:min_hist_records+l_new-1][pred_cols_m].copy()
                        T0 = df_T0.to_dict('list')
                        for key in df_T1.keys():
                            T1[key] = np.array(T1[key])
                        for key in df_T0.keys():
                            T0[key] = np.array(T0[key])
                        
                        result = model.predict(T0, T1)
                        data['Anomaly'] = model.to_ints(result) 
                        data['AnomalyProb'] = data['Anomaly']
                        
                        data.sort_values(by=['ID'], ascending=True, ignore_index=True, inplace=True)
                        cn = GetConnection(db_uri)
                        cn.execute(getSQL(data[['ID', 'SensorSetID','ModelID', 'Anomaly', 'AnomalyProb', 'ModelParameters', 'Ensemble_Positives']], 
                                        query_config['insert_sql']))                     
                        cn.commit()
                        cn.close()
                        print('Insert successfull. Date time: ', datetime.now().strftime("%d.%m.%Y %H:%M:%S"),
                            '. Sensorset ID:', sensorset, '. Model ID: ', modelid,
                            '. Last reading DT:', data['ID'].max().strftime("%d.%m.%Y %H:%M:%S"))
                    else:
                        # There is not enough historical records to perform the forecast
                        # Fill the forecast with default values: Anomaly = 0
                        if len(data)<= min_hist_records-l_hist:
                            l = len(data)
                        else:
                            l = min_hist_records - l_hist
                            
                        df_hist = data.tail(l).copy()                    
                        df_hist['Anomaly'] = 0
                        df_hist['AnomalyProb'] = 0.0
                        df_hist['ModelID'] = modelid
                        df_hist['ModelParameters'] = 'Single model at interval t-15 minutes'
                        df_hist['Ensemble_Positives'] = ''
                        cn = GetConnection(db_uri)
                        cn.execute(getSQL(df_hist[['ID', 'SensorSetID','ModelID', 'Anomaly', 'AnomalyProb', 'ModelParameters', 'Ensemble_Positives']], query_config['insert_sql']))                     
                        cn.commit()
                        cn.close()
                        print('Insert successfull. Date time: ', datetime.now().strftime("%d.%m.%Y %H:%M:%S"),
                            '. Sensorset ID:', sensorset, '. Model ID: ', modelid,
                            '. Last reading DT:', df_hist['ID'].max().strftime("%d.%m.%Y %H:%M:%S"))
                else:
                    print('Sensor ret ID: %s. No new records. Date time: %s' % (sensorset, datetime.now().strftime("%d.%m.%Y %H:%M:%S")))
                    stopper += 1
                    
            # *******************************************************************************
            # ENSEMBLE MODEL FORECASTION 
            if stopper_ens == 0:
                cn = GetConnection(db_uri)
                # Read new data (without the forecast)
                data = pd.read_sql(sql.text(sql_get_new_data.format(modelid_ens,
                                                                    str(sensorset), 
                                                                    str(max_new_records))), cn)
                cn.close() 
                data.sort_values(by=[reading_datetime_col], ascending=False, inplace=True, ignore_index=True)
                l_new = len(data)
                if l_new > 0:
                    # New data is available
                    # Get the reading date and time of the first record
                    data[reading_datetime_col] = pd.to_datetime(data[reading_datetime_col])
                    dt = data[reading_datetime_col].min() 
                    # Get the minimal necessary amount of historical data to perform the forecast
                    cn = GetConnection(db_uri)
                    data_hist = pd.read_sql(sql.text(sql_get_historic_data.format(modelid_ens,
                                                                                str(sensorset), 
                                                                                str(dt), 
                                                                                str(min_hist_records_ens))), cn)
                    cn.close() 
                    l_hist = len(data_hist) 
                    if l_hist >= min_hist_records_ens:
                        # There is enough records to perform the forecast
                        # Concatenate historical and new data
                        data_full = pd.concat([data, data_hist], ignore_index=True)
                        # Sort in descending order by the reading date time so the last new reading would be the first one
                        data_full.sort_values(by=[reading_datetime_col], ascending=False, inplace=True, ignore_index=True) 
                        data['Anomaly'] = 0
                        data['AnomalyProb'] = 0.0
                        data['ModelID'] = modelid_ens
                        data['ModelParameters'] = 'Ensemble of 10 models at intervals of t-30, t-60, t-90, t-120, t-150, t-180, t-210, t-240, t-270 and t-300 minutes' 
                        data['Ensemble_Positives'] = None
                        # Read new records in the list
                        df_T1 = data_full.loc[0:l_new-1][pred_cols].copy()
                        T1 = df_T1.to_dict('list')
                        anom_cols = []
                        for i in hist_records:
                            # Calculating the means
                            for c in pred_cols:
                                data_full[f'{c}_m'] = data_full[c].rolling(window=i).mean()
                            # Read historical records to compare with
                            df_T0 = data_full.loc[i:i+l_new-1][pred_cols_m].copy()
                            T0 = df_T0.to_dict('list')
                            for key in df_T1.keys():
                                T1[key] = np.array(T1[key])
                            for key in df_T0.keys():
                                T0[key] = np.array(T0[key])
                            
                            model = XGBPredictor(PATH=model_config['path'], specify_anomaly=False)
                            result = model.predict(T0, T1)
                            data[f'Anomaly{i}'] = model.to_ints(result)
                            anom_cols.append(f'Anomaly{i}')
                        
                        data.sort_values(by=['ID'], ascending=True, ignore_index=True, inplace=True)
                        data[['Anomaly', 'AnomalyProb', 'Ensemble_Positives']] = make_decision(data[anom_cols].copy(), hist_records)
                        cn = GetConnection(db_uri)
                        cn.execute(getSQL(data[['ID', 'SensorSetID','ModelID', 'Anomaly', 'AnomalyProb', 'ModelParameters', 'Ensemble_Positives']], 
                                        query_config['insert_sql']))                     
                        cn.commit()
                        cn.close()
                        print('Insert successfull. Date time: ', datetime.now().strftime("%d.%m.%Y %H:%M:%S"),
                            '. Sensorset ID:', sensorset, '. Model ID: ', modelid_ens,
                            '. Last reading DT:', data['ID'].max().strftime("%d.%m.%Y %H:%M:%S"))
                    else:
                        # There is not enough historical records to perform the forecast
                        # Fill the forecast with default values: Anomaly = 0
                        if len(data)<= min_hist_records_ens-l_hist:
                            l = len(data)
                        else:
                            l = min_hist_records_ens - l_hist
                            
                        df_hist = data.tail(l).copy()                    
                        df_hist['Anomaly'] = 0
                        df_hist['AnomalyProb'] = 0.0
                        df_hist['ModelID'] = modelid_ens
                        df_hist['ModelParameters'] = 'Ensemble of 10 models at intervals of t-30, t-60, t-90, t-120, t-150, t-180, t-210, t-240, t-270 and t-300 minutes'
                        df_hist['Ensemble_Positives'] = ''
                        cn = GetConnection(db_uri)
                        cn.execute(getSQL(df_hist[['ID', 'SensorSetID','ModelID', 'Anomaly', 'AnomalyProb', 'ModelParameters', 'Ensemble_Positives']], query_config['insert_sql']))                     
                        cn.commit()
                        cn.close()
                        print('Insert successfull. Date time: ', datetime.now().strftime("%d.%m.%Y %H:%M:%S"),
                            '. Sensorset ID:', sensorset, '. Model ID: ', modelid_ens,
                            '. Last reading DT:',df_hist['ID'].max().strftime("%d.%m.%Y %H:%M:%S"))
                else:
                    print('Sensor ret ID: %s. No new records. Date time: %s' % (sensorset, datetime.now().strftime("%d.%m.%Y %H:%M:%S")))
                    stopper_ens += 1
    
        except Error as error:
            print("Failed to insert forecast. Sensor set ID: {}. Error: {}".format(sensorset, error))

        except KeyError as error:
            print("An error occured during the execution of algorithms. Error: {}".format(error))
            
        finally:
            cn.commit()
            cn.close()
    
    
if __name__ == '__main__':      
    get_forecast(1)
    
    